# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Equipment (mechanical, lighting, etc..), and Collection classes."""

import json
from enum import Enum
from pathlib import Path

from pydantic import BaseModel, PrivateAttr


class PhAdorbEquipmentType(str, Enum):
    MECHANICAL = "Mechanical"
    HOT_WATER = "Hot Water"
    APPLIANCE = "Appliance"
    LIGHTS = "Lights"
    PV_ARRAY = "PV Array"
    BATTERY = "Battery"


class PhAdorbEquipment(BaseModel):
    """A single piece of Equipment."""

    name: str
    equipment_type: PhAdorbEquipmentType
    cost: float
    lifetime_years: int
    labor_fraction: float

    @property
    def material_fraction(self) -> float:
        return 1.0 - self.labor_fraction

    def duplicate(self) -> "PhAdorbEquipment":
        return PhAdorbEquipment(
            name=self.name,
            equipment_type=self.equipment_type,
            cost=self.cost,
            lifetime_years=self.lifetime_years,
            labor_fraction=self.labor_fraction,
        )

    def __copy__(self) -> "PhAdorbEquipment":
        return self.duplicate()


class PhAdorbEquipmentCollection(BaseModel):
    """A collection of Equipment."""

    _equipment: dict[str, PhAdorbEquipment] = PrivateAttr(default_factory=dict)

    def add_equipment(self, _ph_adorb_equipment: PhAdorbEquipment) -> None:
        self._equipment[_ph_adorb_equipment.name] = _ph_adorb_equipment

    def get_equipment(self, key: str) -> PhAdorbEquipment:
        return self._equipment[key]

    def keys(self) -> list[str]:
        return [k for k, v in sorted(self._equipment.items(), key=lambda x: x[1].name)]

    def values(self) -> list[PhAdorbEquipment]:
        return list(sorted(self._equipment.values(), key=lambda x: x.name))

    def __iter__(self):
        return iter(sorted(self._equipment.values(), key=lambda x: x.name))

    def __contains__(self, key: str | PhAdorbEquipment) -> bool:
        if isinstance(key, PhAdorbEquipment):
            return key in self._equipment.values()
        return key in self._equipment

    def __len__(self) -> int:
        return len(self._equipment)


def write_equipment_to_json_file(_file_path: Path, equipment: dict[str, PhAdorbEquipment]) -> None:
    """Write all of the Equipment-Types to a JSON file."""
    with open(_file_path, "w") as json_file:
        json.dump([_.dict() for _ in equipment.values()], json_file, indent=4)


def load_equipment_from_json_file(_file_path: Path) -> dict[str, PhAdorbEquipment]:
    """Load all of the Equipment-Types from a JSON file."""
    with open(_file_path, "r") as json_file:
        all_equipment = (PhAdorbEquipment(**item) for item in json.load(json_file))
        return {_.name: _ for _ in all_equipment}



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""A script to generate an ADORB Costs graph from a CSV file.

This script is called from the command line with the following arguments:
    * [1] (str): The path to the CSV file to read in.
    * [2] (str): The path to the folder to save the graphs to.
"""

from collections import namedtuple
import os
import sys

from pathlib import Path
import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio


class InputFileError(Exception):
    def __init__(self, path) -> None:
        self.msg = f"\nCannot find the specified HBJSON file:'{path}'"
        super().__init__(self.msg)


Filepaths = namedtuple(
    "Filepaths",
    [
        "csv",
        "output",
    ],
)


def resolve_paths(_args: list[str]) -> Filepaths:
    """Sort out the file input and output paths. Make the output directory if needed.

    Arguments:
    ----------
        * _args (list[str]): sys.args list of input arguments.

    Returns:
    --------
        * Filepaths
    """

    assert len(_args) == 3, "Error: Incorrect number of arguments. Expected 2, got {}.".format(len(_args))

    # -----------------------------------------------------------------------------------
    # -- The HBJSON input file.
    csv_source_filepath = Path(_args[1])
    if not csv_source_filepath.exists():
        raise InputFileError(csv_source_filepath)

    # -----------------------------------------------------------------------------------
    # -- Graph output folder:
    output_filepath = Path(_args[2])
    if not output_filepath.parent.exists():
        print(f"\t>> Creating the directory: {output_filepath.parent}")
        os.mkdir(output_filepath.parent)

    # -- Make sure it is an HTML output
    output_filepath_ = output_filepath.parent / f"{output_filepath.stem}.html"

    return Filepaths(csv_source_filepath, output_filepath_)


if __name__ == "__main__":
    print("- " * 50)
    print(f"\t>> Using Python: {sys.version}")
    print(f"\t>> Running the script: '{__file__.split('/')[-1]}'")
    print(f"\t>> With the arguments:")
    print("\n".join([f"\t\t{i} | {a}" for i, a in enumerate(sys.argv)]))

    # --- Input / Output file Path
    # -------------------------------------------------------------------------
    print("\t>> Resolving file paths...")
    file_paths = resolve_paths(sys.argv)
    print(f"\t>> Source CSV File: '{file_paths.csv}'")
    print(f"\t>> Target Folder: '{file_paths.output}'")

    df = pd.read_csv(file_paths.csv)

    # -------------------------------------------------------------------------
    # TODO:...
    fig = go.Figure()
    fig.update_layout(title="test")
    # Add each category as a trace
    for column in df.columns:
        fig.add_trace(
            go.Scatter(
                x=df.index, y=df[column], mode="lines", stackgroup="one", name=column  # Creates stacking behavior
            )
        )

    with open(file_paths.output, "w") as f:
        f.write(pio.to_html(fig, full_html=False, include_plotlyjs="cdn"))

    print("\t>> Done generating the ADORB Cost Graphs.")
    print("- " * 50)



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""A simple 'YearlyCost object to store annual costs."""

from dataclasses import dataclass


@dataclass
class YearlyCost:
    """A single Yearly Cost for a building design."""

    cost: float
    year: int
    description: str = ""

    def __repr__(self) -> str:
        return f"YearlyCost(cost={self.cost :.1f}, year={self.year}, description={self.description})"


@dataclass
class YearlyKgCO2:
    """A single Yearly CO2 Emissions for a building design"""

    kg_CO2: float
    year: int
    description: str = ""

    def __repr__(self) -> str:
        return f"YearlyKgCO2(kg_CO2={self.kg_CO2 :.1f}, year={self.year}, description={self.description})"


@dataclass
class YearlyPresentValueFactor:
    """A single Yearly Present Value Factor for a building design."""

    factor: float
    year: int

    def __repr__(self) -> str:
        return f"YearlyPresentValueFactor(pv_factor={self.factor :.3f}, year={self.year})"



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Fuel Types and cost related data."""

from enum import Enum

from pydantic import BaseModel


class PhAdorbFuelType(str, Enum):
    ELECTRICITY = "Electricity"
    NATURAL_GAS = "Natural Gas"


class PhAdorbFuel(BaseModel):
    fuel_type: PhAdorbFuelType
    purchase_price_per_kwh: float
    sale_price_per_kwh: float
    annual_base_price: float
    used: bool = True

    @property
    def name(self) -> str:
        return self.fuel_type.value



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Assembly (wall, floor, etc..) Constructions, and Collection classes."""

import json
from pathlib import Path

from pydantic import BaseModel, PrivateAttr


class PhAdorbConstruction(BaseModel):
    """A single Construction."""

    display_name: str
    identifier: str

    CO2_kg_per_m2: float
    cost_per_m2: float
    lifetime_years: int
    labor_fraction: float
    area_m2: float = 0.0

    @property
    def quantity_ft2(self) -> float:
        return self.area_m2 * 10.7639

    def set_quantity_ft2(self, _value: float) -> None:
        self.area_m2 = _value / 10.7639

    @property
    def cost(self) -> float:
        """Total Cost (quantity * cost_per_m2)."""
        return self.area_m2 * self.cost_per_m2

    @property
    def CO2_kg(self) -> float:
        """Total CO2 (quantity * CO2_kg_per_m2)."""
        return self.area_m2 * self.CO2_kg_per_m2

    @property
    def material_fraction(self) -> float:
        return 1.0 - self.labor_fraction

    def duplicate(self) -> "PhAdorbConstruction":
        return PhAdorbConstruction(
            display_name=self.display_name,
            identifier=self.identifier,
            CO2_kg_per_m2=self.CO2_kg_per_m2,
            cost_per_m2=self.cost_per_m2,
            lifetime_years=self.lifetime_years,
            labor_fraction=self.labor_fraction,
        )

    def __copy__(self) -> "PhAdorbConstruction":
        return self.duplicate()


class PhAdorbConstructionCollection(BaseModel):
    """A collection of Constructions."""

    _constructions: dict[str, PhAdorbConstruction] = PrivateAttr(default_factory=dict)

    def add_construction(self, _construction: PhAdorbConstruction) -> None:
        self._constructions[_construction.display_name] = _construction

    def get_construction(self, key: str) -> PhAdorbConstruction:
        return self._constructions[key]

    def keys(self) -> list[str]:
        return [k for k, v in sorted(self._constructions.items(), key=lambda x: x[1].display_name)]

    def values(self) -> list[PhAdorbConstruction]:
        return list(sorted(self._constructions.values(), key=lambda x: x.display_name))

    def __iter__(self):
        return iter(sorted(self._constructions.values(), key=lambda x: x.display_name))

    def __len__(self) -> int:
        return len(self._constructions)

    def __contains__(self, key: str | PhAdorbConstruction) -> bool:
        if isinstance(key, PhAdorbConstruction):
            return key in self._constructions.values()
        return key in self._constructions

    def set_constructions_ft2_quantities(self, _construction_quantities_ft2: dict[str, float]) -> None:
        """Set the quantity (ft2) of each Construction.

        Args:
            * _construction_quantities (dict[str, float]): A dictionary of Construction
                names, with their quantities (ft2).
        """

        def _clean_name(n: str) -> str:
            return n.upper().replace(" ", "_")

        _construction_quantities_ft2 = {_clean_name(k): v for k, v in _construction_quantities_ft2.items()}

        old_constructions = list(self._constructions.values())
        self._constructions = {}
        for construction in old_constructions:
            new_construction = construction.duplicate()
            new_construction.set_quantity_ft2(_construction_quantities_ft2[_clean_name(construction.display_name)])
            self.add_construction(new_construction)


def write_constructions_to_json_file(_file_path: Path, _constructions: dict[str, PhAdorbConstruction]) -> None:
    """Write all of the Construction-Types to a JSON file."""
    with open(_file_path, "w") as json_file:
        json.dump([_.dict() for _ in _constructions.values()], json_file, indent=4)


def load_constructions_from_json_file(_file_path: Path) -> dict[str, PhAdorbConstruction]:
    """Load all of the Construction-Types from a JSON file."""
    with open(_file_path, "r") as json_file:
        all_constructions = (PhAdorbConstruction(**item) for item in json.load(json_file))
        return {_.display_name: _ for _ in all_constructions}



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Electricity Grid Region with Hourly CO2 Emissions Factors."""

import json
from pathlib import Path

import pandas as pd
from pydantic import BaseModel, Field


class PhAdorbGridRegion(BaseModel):
    """Regional CO2 Emissions Factors for single Electricity Grid Region."""

    region_code: str
    region_name: str
    description: str
    hourly_CO2_factors: dict[int, list[float]] = Field(default_factory=dict)

    def get_CO2_factors_as_df(self) -> pd.DataFrame:
        """
        Returns the CO2 factors as a pandas DataFrame.

        Column=years, rows=hours of the year.
        | Hour | 2023  | 2024  | 2025  |
        |------|-------|-------|-------|
        | 0    | 460.1 | 460.1 | 434.1 |
        | 1    | 469.3 | 475.3 | 445.2 |
        | ...  | ...   | ...   | ...   |
        | 8759 | 460.1 | 460.1 | 434.1 |
        """
        return pd.DataFrame(self.hourly_CO2_factors)


def write_CO2_factors_to_json_file(_file_path: Path, _grid_region: PhAdorbGridRegion):
    """Write the CO2 factors for the grid-region to a JSON file."""
    with open(_file_path, "w") as json_file:
        json.dump(_grid_region.dict(), json_file, indent=4)


def load_CO2_factors_from_json_file(_file_path: Path) -> PhAdorbGridRegion:
    """Load the CO2 factors for the grid-region from a JSON file."""
    with open(_file_path, "r") as json_file:
        return PhAdorbGridRegion(**json.load(json_file))



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Country-level CO2 Emissions data per US-Dollar (kg-CO2/$USD)."""

import json
from pathlib import Path

from pydantic import BaseModel


class PhAdorbNationalEmissions(BaseModel):
    """National Emissions Data."""

    country_name: str
    us_trading_rank: int
    GDP_million_USD: float
    CO2_MT: float
    kg_CO2_per_USD: float


def write_national_emissions_to_json_file(_file_path: Path, _emissions: dict[str, PhAdorbNationalEmissions]) -> None:
    """Write all of the National Emissions data to a JSON file."""
    with open(_file_path, "w") as json_file:
        json.dump([_.dict() for _ in _emissions.values()], json_file, indent=4)


def load_national_emissions_from_json_file(_file_path: Path) -> dict[str, PhAdorbNationalEmissions]:
    """Load all of the National Emissions data from a JSON file."""
    with open(_file_path, "r") as json_file:
        all_emissions = (PhAdorbNationalEmissions(**item) for item in json.load(json_file))
        return {_.country_name: _ for _ in all_emissions}



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""A Building Variant with all of its relevant data, and related functions."""

from pathlib import Path
import logging

import pandas as pd
from pydantic import BaseModel, Field
from ph_units.unit_type import Unit

from ph_adorb import adorb_cost
from ph_adorb.constructions import PhAdorbConstructionCollection
from ph_adorb.equipment import PhAdorbEquipmentCollection
from ph_adorb.fuel import PhAdorbFuel
from ph_adorb.grid_region import PhAdorbGridRegion
from ph_adorb.measures import PhAdorbCO2MeasureCollection
from ph_adorb.national_emissions import PhAdorbNationalEmissions
from ph_adorb.tables.variant import (
    preview_hourly_electric_and_CO2,
    preview_yearly_energy_and_CO2,
    preview_variant_co2_measures,
    preview_variant_constructions,
    preview_variant_equipment,
    preview_yearly_embodied_CO2_costs,
    preview_yearly_embodied_kgCO2,
    preview_yearly_install_costs,
)
from ph_adorb.yearly_values import YearlyCost, YearlyKgCO2

logger = logging.getLogger(__name__)

# ---------------------------------------------------------------------------------------


class PhAdorbVariant(BaseModel):
    """A single Variant of a building design."""

    name: str
    total_purchased_gas_kwh: float
    hourly_purchased_electricity_kwh: list[float]
    total_sold_electricity_kwh: float
    peak_electric_usage_W: float
    electricity: PhAdorbFuel
    gas: PhAdorbFuel
    grid_region: PhAdorbGridRegion
    national_emissions: PhAdorbNationalEmissions
    analysis_duration: int
    envelope_labor_cost_fraction: float

    measure_collection: PhAdorbCO2MeasureCollection = Field(default_factory=PhAdorbCO2MeasureCollection)
    construction_collection: PhAdorbConstructionCollection = Field(default_factory=PhAdorbConstructionCollection)
    equipment_collection: PhAdorbEquipmentCollection = Field(default_factory=PhAdorbEquipmentCollection)

    price_of_carbon: float = 0.25

    @property
    def total_purchased_electricity_kwh(self) -> float:
        """Return the total annual purchased electricity in KWH."""
        return sum(self.hourly_purchased_electricity_kwh)

    class Config:
        arbitrary_types_allowed = True

    @property
    def all_carbon_measures(self) -> PhAdorbCO2MeasureCollection:
        """Return a collection of all the Carbon Measures."""
        return self.measure_collection

    @property
    def performance_measure_collection(self) -> PhAdorbCO2MeasureCollection:
        """Return a collection of only the Performance Carbon Measures."""
        return self.measure_collection.performance_measures

    @property
    def nonperformance_carbon_measures(self) -> PhAdorbCO2MeasureCollection:
        """Return a collection of only the Non-Performance Carbon Measures."""
        return self.measure_collection.nonperformance_measures


# ---------------------------------------------------------------------------------------


def calc_annual_total_electric_cost(
    _purchased_electricity_kwh: float,
    _sold_electricity_kwh: float,
    _electric_purchase_price_per_kwh: float,
    _electric_sell_price_per_kwh: float,
    _electric_annual_base_price: float,
) -> float:
    """Return the total annual electricity cost for the building."""
    logger.info("calc_annual_total_electric_cost()")

    total_purchased_electric_cost = _purchased_electricity_kwh * _electric_purchase_price_per_kwh
    total_sold_electric_cost = _sold_electricity_kwh * _electric_sell_price_per_kwh
    total_annual_electric_cost = total_purchased_electric_cost - total_sold_electric_cost + _electric_annual_base_price

    # ------------------------------------------------------------------------------------------------------------------
    logger.debug(
        f"Electric Purchased: {_purchased_electricity_kwh :,.0f}kWh * ${_electric_purchase_price_per_kwh :,.2f}/kWh = ${total_purchased_electric_cost :,.0f}"
    )
    logger.debug(
        f"Electric Sold: {_sold_electricity_kwh :,.0f}kWh * ${_electric_sell_price_per_kwh :,.2f}/kWh = ${total_sold_electric_cost :,.0f}"
    )
    logger.debug(
        f"Electric Net Cost: ${total_purchased_electric_cost :,.0f} - ${total_sold_electric_cost :,.0f} + ${_electric_annual_base_price :,.0f} = ${total_annual_electric_cost :,.0f}"
    )

    return total_annual_electric_cost


def calc_annual_hourly_electric_CO2(
    _hourly_purchased_electricity_kwh: list[float], _grid_region: PhAdorbGridRegion
) -> list[float]:
    """Return a list of total annual CO2 emissions for each year from 2023 - 2011 (89 years)."""
    MWH_PER_KWH = 0.001

    # -- Convert the hourly purchased electricity from KWH to a pd.Series as MWH
    hourly_electric_MWH = pd.Series(_hourly_purchased_electricity_kwh) * MWH_PER_KWH

    # Multiply each year's factors by the hourly electric MWH list, and sum the results for each year.
    annual_hourly_electric_CO2: list[float] = (
        (_grid_region.get_CO2_factors_as_df().multiply(hourly_electric_MWH, axis=0)).sum().tolist()
    )
    return annual_hourly_electric_CO2


def calc_annual_total_gas_cost(
    _total_purchased_gas_kwh: float,
    _gas_used: bool,
    _gas_purchase_price_per_kwh: float,
    _gas_annual_base_price: float,
) -> float:
    """Return the total annual gas cost for the building."""
    logger.info("calc_annual_total_gas_cost()")

    if not _gas_used:
        return 0.0

    total_annual_gas_cost = (_total_purchased_gas_kwh * _gas_purchase_price_per_kwh) + _gas_annual_base_price

    logger.debug(
        f"Gas Cost: {_total_purchased_gas_kwh :,.0f}kWh * ${_gas_purchase_price_per_kwh :,.2f}/kWh + ${_gas_annual_base_price :,.0f} = ${total_annual_gas_cost :,.0f}"
    )
    return total_annual_gas_cost


def calc_annual_total_gas_CO2(
    _total_purchased_gas_kwh: float,
    _gas_used: bool,
) -> float:
    logger.info("calc_annual_total_gas_CO2()")

    TONS_CO2_PER_THERM_GAS = 12.7

    if not _gas_used:
        return 0.0

    annual_therms_gas = Unit(_total_purchased_gas_kwh, "KWH").as_a("THERM").value
    annual_tons_gas_CO2 = annual_therms_gas * TONS_CO2_PER_THERM_GAS

    logger.debug(
        f"Gas CO2: {_total_purchased_gas_kwh :,.0f} kWH -> {annual_therms_gas :,.0f} Therms * {TONS_CO2_PER_THERM_GAS} = {annual_tons_gas_CO2 :,.0f} tons CO2"
    )

    return annual_tons_gas_CO2


# ---------------------------------------------------------------------------------------
# -- CO2-Reduction-Measures


def calc_CO2_reduction_measures_yearly_embodied_kgCO2(
    _variant_CO2_measures: PhAdorbCO2MeasureCollection,
    _kg_CO2_per_USD: float,
) -> list[YearlyKgCO2]:
    """Return a list of all the Yearly-Embodied-kgCO2 for all the Variant's CO2-Reduction-Measures."""

    logger.info(f"calc_CO2_reduction_measures_yearly_embodied_kgCO2({len(_variant_CO2_measures)} measures)")

    # TODO: CHANGE TO USE COUNTRY INDEX, 0 for US,
    yearly_embodied_kgCO2_: list[YearlyKgCO2] = []

    for measure in _variant_CO2_measures:
        measure_kgCO2 = measure.cost * _kg_CO2_per_USD
        yearly_embodied_kgCO2_.append(YearlyKgCO2(measure_kgCO2, measure.year, measure.name))

        logger.debug(
            f"CO2 Measure {measure.name} [YR-{measure.year}]: ${measure.cost :,.0f} * {_kg_CO2_per_USD :,.0f} kgCO2/USD = {measure_kgCO2 :,.0f}"
        )

    # TODO: Labor fraction should be subtracted out and have USA EF applied

    return yearly_embodied_kgCO2_


def calc_CO2_reduction_measures_yearly_embodied_CO2_cost(
    _yearly_embodied_kgCO2_: list[YearlyKgCO2], _USD_per_kgCO2=0.25
) -> list[YearlyCost]:
    """Return a list of all the Yearly-Embodied-CO2-Costs for all the Variant's CO2-Reduction-Measures."""
    logger.info("calc_CO2_reduction_measures_yearly_embodied_CO2_cost()")

    yearly_CO2_costs_ = []
    for yearly_kgCO2 in _yearly_embodied_kgCO2_:
        CO2_cost = yearly_kgCO2.kg_CO2 * _USD_per_kgCO2
        logger.debug(
            f"CO2 Measure {yearly_kgCO2.description} [YR-{yearly_kgCO2.year}]: {yearly_kgCO2.kg_CO2 :,.0f} kgCO2 * ${_USD_per_kgCO2 :,.2f}/kgCO2 = ${CO2_cost :,.0f}"
        )

        yearly_CO2_costs_.append(YearlyCost(CO2_cost, yearly_kgCO2.year, yearly_kgCO2.description))

    return yearly_CO2_costs_


# TODO: Do  we need this? What is the '_envelope_labor_cost_fraction' doing here?
# def get_first_cost_embodied_CO2_cost(
#     _envelope_labor_cost_fraction: float,
#     _first_costs: YearlyCost,
#     _kg_CO2_per_USD: float,
# ) -> YearlyCost:
#     """Return the first cost of the embodied CO2 for the Carbon Measures."""

#     material_fraction: float = 1.0 - _envelope_labor_cost_fraction
#     material_first_cost = _first_costs.cost * material_fraction * _kg_CO2_per_USD
#     labor_first_cost = _first_costs.cost * _envelope_labor_cost_fraction * _kg_CO2_per_USD
#     total_first_cost = material_first_cost + labor_first_cost

#     return YearlyCost(total_first_cost, 0, "First Cost")


def calc_CO2_reduction_measures_yearly_install_costs(
    _variant_CO2_measures: PhAdorbCO2MeasureCollection,
) -> list[YearlyCost]:
    """Return a list of all the Yearly-Install-Costs (labor + material) for all the Variant's CO2-Reduction-Measures."""
    logger.info("calc_CO2_reduction_measures_yearly_install_costs()")

    return [YearlyCost(measure.cost, measure.year, measure.name) for measure in _variant_CO2_measures]


# ---------------------------------------------------------------------------------------
# -- Constructions


def calc_constructions_yearly_embodied_kgCO2(
    _construction_collection: PhAdorbConstructionCollection, _analysis_duration, _kg_CO2_per_USD
) -> list[YearlyKgCO2]:
    """Return a list of all the Yearly-Embodied-CO2-Costs for all the Variant's Construction Materials."""
    logger.info("calc_constructions_yearly_embodied_kgCO2()")

    yearly_embodied_kgCO2_: list[YearlyKgCO2] = []
    for const in _construction_collection:
        const_material_dollar_cost: float = const.cost * const.material_fraction
        const_material_embodied_kgCO2: float = const_material_dollar_cost * _kg_CO2_per_USD

        logger.debug(
            f"Construction {const.display_name}: ${const_material_dollar_cost :,.0f} * {_kg_CO2_per_USD :,.2f} kgCO2/USD = {const_material_embodied_kgCO2 :,.0f} kgCO2"
        )

        for year in range(0, _analysis_duration + 1, const.lifetime_years or (_analysis_duration + 1)):
            logger.debug(
                f"Adding Construction {const.display_name} Embodied CO2 [lifetime={const.lifetime_years}yrs] {const_material_embodied_kgCO2 :,.0f} kgCO2 for year-{year}"
            )
            yearly_embodied_kgCO2_.append(YearlyKgCO2(const_material_embodied_kgCO2, year, const.display_name))

    return yearly_embodied_kgCO2_


def calc_constructions_yearly_embodied_CO2_cost(
    _yearly_embodied_kgCO2_: list[YearlyKgCO2], _USD_per_kgCO2=0.25
) -> list[YearlyCost]:
    """Return a list of all the Yearly-Embodied-CO2-Costs for all the Variant's Construction Materials."""
    logger.info("calc_constructions_yearly_embodied_CO2_cost()")

    yearly_embodied_CO2_ = []
    for yearly_kgCO2 in _yearly_embodied_kgCO2_:
        CO2_cost = yearly_kgCO2.kg_CO2 * _USD_per_kgCO2
        logger.debug(
            f"Construction {yearly_kgCO2.description} Embodied CO2-Cost [YR-{yearly_kgCO2.year}]: {yearly_kgCO2.kg_CO2 :,.0f} kgCO2 * ${_USD_per_kgCO2 :,.2f}/kgCO2 = ${CO2_cost :,.0f}"
        )

        yearly_embodied_CO2_.append(YearlyCost(CO2_cost, yearly_kgCO2.year, yearly_kgCO2.description))

    return yearly_embodied_CO2_


def calc_constructions_yearly_install_costs(
    _construction_collection: PhAdorbConstructionCollection,
    _analysis_duration,
) -> list[YearlyCost]:
    """Return a list of all the Yearly-Install-Costs (labor + material) for all the Variant's Constructions."""
    logger.info("calc_constructions_yearly_install_costs()")

    yearly_install_costs_ = []
    for const in _construction_collection:
        for year in range(0, _analysis_duration + 1, const.lifetime_years or (_analysis_duration + 1)):
            logger.debug(
                f"Adding Construction {const.display_name} Install Cost: [lifetime={const.lifetime_years}yrs] ${const.cost :,.0f} for year-{year}"
            )
            yearly_install_costs_.append(YearlyCost(const.cost, year, const.display_name))
    return yearly_install_costs_


# ---------------------------------------------------------------------------------------
# -- Mechanical Equipment & Appliances


def calc_equipment_yearly_embodied_kgCO2_(
    _equipment_collection: PhAdorbEquipmentCollection, _analysis_duration, _kg_CO2_per_USD
) -> list[YearlyKgCO2]:
    """Return a list of all the Yearly-Embodied-kgCO2 for all the Variant's Equipment."""
    logger.info("calc_equipment_yearly_embodied_kgCO2_()")

    yearly_embodied_kgCO2_: list[YearlyKgCO2] = []
    for equip in _equipment_collection:
        equip_material_cost: float = equip.cost * equip.material_fraction
        equip_material_embodied_CO2_cost: float = equip_material_cost * _kg_CO2_per_USD

        logger.debug(
            f"Equipment {equip.name}: ${equip_material_cost :,.0f} * {_kg_CO2_per_USD :,.2f} kgCO2/USD = {equip_material_embodied_CO2_cost :,.0f} kgCO2"
        )

        for year in range(0, _analysis_duration + 1, equip.lifetime_years or (_analysis_duration + 1)):
            logger.debug(
                f"Adding Equipment {equip.name} Embodied CO2 [lifetime={equip.lifetime_years}yrs] {equip_material_embodied_CO2_cost :,.0f} kgCO2 for year-{year}"
            )
            yearly_embodied_kgCO2_.append(YearlyKgCO2(equip_material_embodied_CO2_cost, year, equip.name))

    return yearly_embodied_kgCO2_


def calc_equipment_yearly_embodied_CO2_cost(
    _yearly_embodied_kgCO2_: list[YearlyKgCO2], _USD_per_kgCO2=0.25
) -> list[YearlyCost]:
    """Return a list of all the Yearly-Embodied-CO2-Costs for all the Variant's Equipment."""
    logger.info("calc_equipment_yearly_embodied_CO2_cost()")

    yearly_embodied_CO2_cost_ = []
    for yearly_kgCO2 in _yearly_embodied_kgCO2_:
        CO2_cost = yearly_kgCO2.kg_CO2 * _USD_per_kgCO2
        logger.debug(
            f"Equipment {yearly_kgCO2.description} Embodied CO2-Cost [YR-{yearly_kgCO2.year}]: {yearly_kgCO2.kg_CO2 :,.0f} kgCO2 * ${_USD_per_kgCO2 :,.2f}/kgCO2 = ${CO2_cost :,.0f}"
        )

        yearly_embodied_CO2_cost_.append(YearlyCost(CO2_cost, yearly_kgCO2.year, yearly_kgCO2.description))

    return yearly_embodied_CO2_cost_


def calc_equipment_yearly_install_costs(
    _equipment_collection: PhAdorbEquipmentCollection,
    _analysis_duration,
) -> list[YearlyCost]:
    """Return a list of all the Yearly-Install-Costs (labor + material) for all the Variant's Equipment."""
    logger.info("calc_equipment_yearly_install_costs()")

    yearly_install_costs_ = []
    for equip in _equipment_collection:
        for year in range(0, _analysis_duration + 1, equip.lifetime_years or (_analysis_duration + 1)):
            logger.debug(
                f"Adding Equipment {equip.name} Install Cost: [lifetime={equip.lifetime_years}yrs] ${equip.cost :,.0f} for year-{year}"
            )
            yearly_install_costs_.append(YearlyCost(equip.cost, year, equip.name))

    return yearly_install_costs_


# ---------------------------------------------------------------------------------------


def calc_variant_yearly_ADORB_costs(_variant: PhAdorbVariant, _output_tables_path: Path | None = None) -> pd.DataFrame:
    """Return a DataFrame with the Variant's yearly ADORB costs for each year of the analysis duration."""
    logger.info("calc_variant_yearly_ADORB_costs()")

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    # -- Electric: Annual Costs, Annual CO2
    annual_total_cost_electric = calc_annual_total_electric_cost(
        _variant.total_purchased_electricity_kwh,
        _variant.total_sold_electricity_kwh,
        _variant.electricity.purchase_price_per_kwh,
        _variant.electricity.sale_price_per_kwh,
        _variant.electricity.annual_base_price,
    )
    future_annual_total_CO2_electric = calc_annual_hourly_electric_CO2(
        _variant.hourly_purchased_electricity_kwh,
        _variant.grid_region,
    )

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    # -- Gas: Annual Costs, Annual CO2
    annual_total_cost_gas = calc_annual_total_gas_cost(
        _variant.total_purchased_gas_kwh,
        _variant.gas.used,
        _variant.gas.purchase_price_per_kwh,
        _variant.gas.annual_base_price,
    )
    annual_total_CO2_gas = calc_annual_total_gas_CO2(
        _variant.total_purchased_gas_kwh,
        _variant.gas.used,
    )

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    # -- CO2-REDUCTION-MEASURES Costs
    carbon_measure_yearly_embodied_kgCO2 = calc_CO2_reduction_measures_yearly_embodied_kgCO2(
        _variant.all_carbon_measures,
        _variant.national_emissions.kg_CO2_per_USD,
    )
    carbon_measure_yearly_embodied_CO2_costs = calc_CO2_reduction_measures_yearly_embodied_CO2_cost(
        carbon_measure_yearly_embodied_kgCO2,
        _variant.price_of_carbon,
    )
    carbon_measure_yearly_install_costs = calc_CO2_reduction_measures_yearly_install_costs(_variant.all_carbon_measures)

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    # -- CONSTRUCTIONS Costs
    construction_yearly_embodied_kgCO2 = calc_constructions_yearly_embodied_kgCO2(
        _variant.construction_collection,
        _variant.analysis_duration,
        _variant.national_emissions.kg_CO2_per_USD,
    )
    construction_yearly_embodied_CO2_costs = calc_constructions_yearly_embodied_CO2_cost(
        construction_yearly_embodied_kgCO2, _variant.price_of_carbon
    )
    construction_yearly_install_costs = calc_constructions_yearly_install_costs(
        _variant.construction_collection,
        _variant.analysis_duration,
    )

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    # -- HVAC EQUIPMENT AND APPLIANCE Costs
    equipment_yearly_embodied_kgCO2_ = calc_equipment_yearly_embodied_kgCO2_(
        _variant.equipment_collection,
        _variant.analysis_duration,
        _variant.national_emissions.kg_CO2_per_USD,
    )
    equipment_yearly_embodied_CO2_costs = calc_equipment_yearly_embodied_CO2_cost(
        equipment_yearly_embodied_kgCO2_,
        _variant.price_of_carbon,
    )
    equipment_yearly_install_costs = calc_equipment_yearly_install_costs(
        _variant.equipment_collection,
        _variant.analysis_duration,
    )

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    all_yearly_install_costs = (
        carbon_measure_yearly_install_costs + construction_yearly_install_costs + equipment_yearly_install_costs
    )
    all_yearly_embodied_kgCO2 = (
        carbon_measure_yearly_embodied_kgCO2 + construction_yearly_embodied_kgCO2 + equipment_yearly_embodied_kgCO2_
    )
    all_yearly_embodied_kgCO2_costs = (
        carbon_measure_yearly_embodied_CO2_costs
        + construction_yearly_embodied_CO2_costs
        + equipment_yearly_embodied_CO2_costs
    )

    if _output_tables_path:
        logger.info(f"Saving ADORB Tables to: {_output_tables_path}")
        preview_hourly_electric_and_CO2(
            _variant.hourly_purchased_electricity_kwh,
            _variant.grid_region.hourly_CO2_factors,
            _output_tables_path,
        )
        preview_yearly_energy_and_CO2(
            _variant.total_purchased_electricity_kwh,
            future_annual_total_CO2_electric,
            _variant.total_purchased_gas_kwh,
            annual_total_CO2_gas,
            _output_tables_path,
        )
        preview_variant_co2_measures(_variant.measure_collection, _output_tables_path)
        preview_variant_constructions(_variant.construction_collection, _output_tables_path)
        preview_variant_equipment(_variant.equipment_collection, _output_tables_path)
        preview_yearly_install_costs(all_yearly_install_costs, _output_tables_path)
        preview_yearly_embodied_kgCO2(all_yearly_embodied_kgCO2, _output_tables_path)
        preview_yearly_embodied_CO2_costs(all_yearly_embodied_kgCO2_costs, _output_tables_path)

    # -----------------------------------------------------------------------------------
    # -----------------------------------------------------------------------------------
    # -- Compute and return the ADORB costs DataFrame
    return adorb_cost.calculate_annual_ADORB_costs(
        _variant.analysis_duration,
        annual_total_cost_electric,
        annual_total_cost_gas,
        future_annual_total_CO2_electric,
        annual_total_CO2_gas,
        all_yearly_install_costs,
        all_yearly_embodied_kgCO2_costs,
        _variant.peak_electric_usage_W,
        _variant.price_of_carbon,
    )


def calc_variant_cumulative_ADORB_costs(_df: pd.DataFrame) -> pd.DataFrame:
    """Return a DataFrame with the Cumulative ADORB costs for each year of the analysis duration."""
    logger.info("calc_variant_cumulative_ADORB_costs()")

    cumulative_df = _df.cumsum(axis=0)
    return cumulative_df



# -*- coding: utf-8 -*-
# -*- Python Version: 3.10 -*-

"""Functions to preview variant costs in table-format."""

from collections import defaultdict
from pathlib import Path

from rich.console import Console
from rich.table import Table

from ph_adorb.measures import PhAdorbCO2MeasureCollection
from ph_adorb.constructions import PhAdorbConstructionCollection
from ph_adorb.equipment import PhAdorbEquipmentCollection
from ph_adorb.yearly_values import YearlyCost, YearlyKgCO2


def rich_table_to_html(_tbl: Table) -> str:
    """Convert a rich-Table to an HTML-table.

    Args:
        _tbl (Table): The rich Table to convert.
    Returns:
        str: The HTML table as a string.
    """

    # Extract headers
    headers = [column.header for column in _tbl.columns]

    # Extract row data
    rows = [[column._cells[i] for column in _tbl.columns] for i in range(_tbl.row_count)]

    # Construct HTML table with inline styling
    html_ = '<table style="border-collapse: collapse; width: 100%; font-family: sans-serif;">\n'
    html_ += f'  <caption style="caption-side: top; font-size: 1.5em; font-weight: bold; padding: 10px;">{_tbl.title}</caption>\n'
    html_ += '  <thead>\n    <tr style="background-color: #f2f2f2; font-weight: bold;">\n'
    for header in headers:
        html_ += f'      <th style="padding: 8px; text-align: center;">{header}</th>\n'
    html_ += "    </tr>\n  </thead>\n"
    html_ += "  <tbody>\n"
    for i, row in enumerate(rows):
        background_color = "#ffffff" if i % 2 == 0 else "#f2f2f2"
        html_ += f'    <tr style="background-color: {background_color};">\n'
        for cell in row:
            html_ += f'      <td style="padding: 8px; text-align: center;">{cell}</td>\n'
        html_ += "    </tr>\n"
    html_ += "  </tbody>\n</table>\n"

    return html_


def add_total_row(tbl_: Table) -> None:
    """Add a total row to the table if it contains numeric columns."""
    if tbl_.row_count == 0:
        return

    total_row = ["Total"]
    for col_idx in range(1, len(tbl_.columns)):
        try:
            total = sum(
                float(str(tbl_.columns[col_idx]._cells[row_idx]).replace(",", ""))
                for row_idx in range(tbl_.row_count)
                if str(tbl_.columns[col_idx]._cells[row_idx]).replace(",", "").replace(".", "").isdigit()
            )
            total_row.append(f"{total:,.0f}")
        except ValueError:
            total_row.append("-")

    tbl_.add_row(*total_row, style="bold")


def preview_hourly_electric_and_CO2(
    _hourly_kwh: list[float], _hourly_CO2_factors: dict[int, list[float]], _output_path: Path | None
) -> None:
    # Create the table
    tbl_ = Table(title="Hourly Electric Consumption (kWh) and CO2 Factors (kgCO2/kWh)", show_lines=True)
    tbl_.add_column("Hour", style="cyan", justify="center", min_width=20, no_wrap=True)
    tbl_.add_column("kWh", style="magenta", justify="center")
    for year in _hourly_CO2_factors.keys():
        tbl_.add_column(f"{year}", style="magenta", justify="center")

    # Iterate over the hourly data and add rows to the table
    for i, kwh in enumerate(_hourly_kwh):
        factors_by_year: list[str] = []
        for yearly_factors in _hourly_CO2_factors.values():
            factors_by_year.append(f"{yearly_factors[i]:,.0f}")
        tbl_.add_row(f"{i:04d}", f"{kwh:,.2f}", *factors_by_year)

    add_total_row(tbl_)

    # Output the table to the console or write to a file
    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "hourly_electric_kwh.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_yearly_energy_and_CO2(
    _elec_kwh: float, _elec_CO2_by_future_year: list[float], _gas_kwh: float, _gas_CO2: float, _output_path: Path | None
) -> None:
    # Create the table
    tbl_ = Table(title="Future Annual Energy Consumption and CO2 Emissions", show_lines=True)
    tbl_.add_column("Year", style="cyan", justify="center", min_width=20, no_wrap=True)
    tbl_.add_column("Elec. kWh", style="magenta", justify="center")
    tbl_.add_column("Elec. kgCO2", style="magenta", justify="center")
    tbl_.add_column("Elec. Grid Factor (kgCO2/kWh)", style="magenta", justify="center")
    tbl_.add_column("Gas kWh", style="magenta", justify="center")
    tbl_.add_column("Gas kgCO2", style="magenta", justify="center")
    tbl_.add_column("Gas Grid Factor (kgCO2/kWh)", style="magenta", justify="center")

    # Iterate over the hourly data and add rows to the table
    for i, co2 in enumerate(
        _elec_CO2_by_future_year,
    ):
        try:
            elec_grid_factor = co2 / _elec_kwh
        except ZeroDivisionError:
            elec_grid_factor = 0.0

        try:
            gas_grid_factor = _gas_CO2 / _gas_kwh
        except ZeroDivisionError:
            gas_grid_factor = 0.0

        tbl_.add_row(
            f"{2023+i:02d}",
            f"{_elec_kwh:,.0f}",
            f"{co2:,.0f}",
            f"{elec_grid_factor:,.3f}",
            f"{_gas_kwh:,.0f}",
            f"{_gas_CO2:,.0f}",
            f"{gas_grid_factor:,.3f}",
        )

    add_total_row(tbl_)

    # Output the table to the console or write to a file
    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "annual_electric_kwh_and_CO2.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_variant_co2_measures(
    _co2_measure_collection: PhAdorbCO2MeasureCollection, _output_path: Path | None
) -> None:
    # Create the table
    tbl_ = Table(title="Variant CO2 Reduction Measures", show_lines=True)
    tbl_.add_column("Measure", style="cyan", justify="center", min_width=20, no_wrap=True)
    tbl_.add_column("Type", style="magenta", justify="center")
    tbl_.add_column("Year", style="magenta", justify="center")
    tbl_.add_column("USD", style="magenta", justify="center")
    tbl_.add_column("(kgCO2)", style="magenta", justify="center")
    tbl_.add_column("Country Name", style="magenta", justify="center")
    tbl_.add_column("Labor Fraction [%]", style="magenta", justify="center")

    # Iterate over the CO2 measure collection and add rows to the table
    for measure in _co2_measure_collection:
        tbl_.add_row(
            measure.name,
            measure.measure_type.name,
            f"{measure.year}",
            f"{measure.cost:,.0f}",
            f"{measure.kg_CO2:.0f}" if measure.kg_CO2 is not None else "-",
            f"{measure.country_name}",
            f"{measure.labor_fraction * 100.0 :.0f}",
        )

    add_total_row(tbl_)

    # Output the table to the console or write to a file
    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "co2_measures.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_variant_equipment(_equipment_collection: PhAdorbEquipmentCollection, _output_path: Path | None) -> None:
    """Preview the variant equipment in a table."""
    # Create the table
    tbl_ = Table(title="Variant Equipment", show_lines=True)
    tbl_.add_column("Equipment/Appliance", style="cyan", justify="center", min_width=20, no_wrap=True)
    tbl_.add_column("Type", style="magenta", justify="center")
    tbl_.add_column("USD", style="magenta", justify="center")
    tbl_.add_column("Lifetime (years)", style="magenta", justify="center")
    tbl_.add_column("Labor Fraction [%]", style="magenta", justify="center")

    # Iterate over the equipment collection and add rows to the table
    for equipment in _equipment_collection:
        tbl_.add_row(
            equipment.name,
            equipment.equipment_type.name,
            f"{equipment.cost:,.0f}",
            f"{equipment.lifetime_years:.0f}",
            f"{equipment.labor_fraction * 100.0 :.0f}",
        )

    add_total_row(tbl_)

    # Output the table to the console or write to a file
    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "equipment.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_variant_constructions(
    _construction_collection: PhAdorbConstructionCollection, _output_path: Path | None
) -> None:
    # Create the table
    tbl_ = Table(title="Variant Constructions", show_lines=True)
    tbl_.add_column("Construction", style="cyan", justify="center", min_width=20, no_wrap=True)
    tbl_.add_column("Area (M2)", style="magenta", justify="center")
    tbl_.add_column("USD/M2", style="magenta", justify="center")
    tbl_.add_column("Total USD", style="magenta", justify="center")
    tbl_.add_column("kgCO2 / M2", style="magenta", justify="center")
    tbl_.add_column("Total kgCO2", style="magenta", justify="center")
    tbl_.add_column("Lifetime (years)", style="magenta", justify="center")
    tbl_.add_column("Labor Fraction [%]", style="magenta", justify="center")

    # Iterate over the construction collection and add rows to the table
    for construction in _construction_collection:

        tbl_.add_row(
            construction.display_name,
            f"{construction.area_m2:,.1f}",
            f"{construction.cost_per_m2:,.2f}",
            f"{construction.cost:,.0f}",
            f"{construction.CO2_kg_per_m2:,.2f}",
            f"{construction.CO2_kg:,.0f}",
            f"{construction.lifetime_years:.0f}",
            f"{construction.labor_fraction * 100.0 :.0f}",
        )

    add_total_row(tbl_)

    # Output the table to the console or write to a file
    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "constructions.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_yearly_install_costs(_input: list[YearlyCost], _output_path: Path | None) -> None:
    # Group the data by description
    grouped_data = defaultdict(lambda: defaultdict(float))
    unique_years = set()

    for item in _input:
        grouped_data[item.description][item.year] += item.cost
        unique_years.add(item.year)

    # Sort the years to ensure columns are in order
    sorted_years = sorted(unique_years)

    # Create the table
    tbl_ = Table(title="Variant Install Costs (USD) by Year", show_lines=True)
    tbl_.add_column("Description", style="cyan", justify="center", min_width=20, no_wrap=True)

    for year in sorted_years:
        tbl_.add_column(str(year), style="magenta", justify="center", min_width=8)

    for description, costs in grouped_data.items():
        row = [description] + [
            f"{costs.get(year, 0):,.0f}" if costs.get(year, 0) != 0 else "-" for year in sorted_years
        ]
        tbl_.add_row(*row)

    add_total_row(tbl_)

    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "yearly_install_costs.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_yearly_embodied_kgCO2(_input: list[YearlyKgCO2], _output_path: Path | None) -> None:
    # Group the data by description
    grouped_data = defaultdict(lambda: defaultdict(float))
    unique_years = set()

    for item in _input:
        grouped_data[item.description][item.year] += item.kg_CO2
        unique_years.add(item.year)

    # Sort the years to ensure columns are in order
    sorted_years = sorted(unique_years)

    # Create the table
    tbl_ = Table(title="Variant Embodied-CO2 (kgCO2) by Year", show_lines=True)
    tbl_.add_column("Description", style="cyan", justify="center", min_width=20, no_wrap=True)

    for year in sorted_years:
        tbl_.add_column(str(year), style="magenta", justify="center", min_width=8)

    for description, costs in grouped_data.items():
        row = [description] + [
            f"{costs.get(year, 0):,.0f}" if costs.get(year, 0) != 0 else "-" for year in sorted_years
        ]
        tbl_.add_row(*row)

    add_total_row(tbl_)

    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "yearly_embodied_CO2_kg.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)


def preview_yearly_embodied_CO2_costs(_input: list[YearlyCost], _output_path: Path | None) -> None:
    # Group the data by description
    grouped_data = defaultdict(lambda: defaultdict(float))
    unique_years = set()

    for item in _input:
        grouped_data[item.description][item.year] += item.cost
        unique_years.add(item.year)

    # Sort the years to ensure columns are in order
    sorted_years = sorted(unique_years)

    # Create the table
    tbl_ = Table(title="Variant Embodied-CO2 Costs (USD) by Year.", show_lines=True)
    tbl_.add_column("Description", style="cyan", justify="center", min_width=20, no_wrap=True)

    for year in sorted_years:
        tbl_.add_column(str(year), style="magenta", justify="center", min_width=8)

    for description, costs in grouped_data.items():
        row = [description] + [
            f"{costs.get(year, 0):,.0f}" if costs.get(year, 0) != 0 else "-" for year in sorted_years
        ]
        tbl_.add_row(*row)

    add_total_row(tbl_)

    if _output_path:
        html_table = rich_table_to_html(tbl_)
        with open(Path(_output_path / "yearly_embodied_CO2_costs.html"), "w") as f:
            f.write(html_table)
    else:
        console = Console()
        console.print(tbl_)
